using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using ScannerMng;
using SIP = ScannerMng.SIProperty;	// SIP will be shorthand for ScannerMng.SIProperty

namespace CSSample
{

	public partial class CSSample : Form
	{
		// This is the filename of the DLL providing the "native" API.
		// For the sample, use the DocketPORT 487 DLL.
		private string m_ScannerDllFilename = "DPORT487.dll";

		// This is the scanner name that is passed to SI_OpenInterface 
		// to specify we want to open a DocketPORT 485.
		private string m_ScannerName = "DocketPORT487";

		// The filename of the image file to be written.
		private string m_outFilename = "CSSample.bmp";

		private SIScanner m_scanner;
		private bool m_scannerReady;
		private ProgressDialog m_progressDialog;

		const string m_blackAndWhiteText = "B&W";
		const string m_grayText = "Gray";
		const string m_colorText = "Color";

		public CSSample()
		{
			InitializeComponent();

			m_scannerReady = false;

			try
			{
				m_scanner = new SIScanner(m_ScannerDllFilename);
			}
			catch (ApplicationException e)
			{
				MessageBox.Show(e.Message);
				return;
			}

			SIResult result = m_scanner.SI_OpenInterface(m_ScannerName);

			if (result != SIResult.SIR_SUCCESS)
			{
				string str = "";
				m_scanner.SI_GetLastErrorText(ref str);
				MessageBox.Show(str);
				return;
			}

			m_scannerReady = true;

			PopulateControls();

		}

		// This function will query the scanner for its available scan modes 
		// and resolutions. Then it will add these to the UI controls. 
		// This is not strictly necessary. If you're only developing for one 
		// scanner model and you know the available modes and resolutions,
		// then you can just hard code them. But querying the scanner properties
		// is useful if you want to write code that works for more than one 
		// scanner model.
		//
		private void PopulateControls()
		{
			if (!m_scannerReady)
			{
				return;
			}

			SIProperty prop = new SIProperty();
			SIResult result;

			// Get Scan modes			
			prop.propertyID = SIP.SIP_SCAN_MODE;
			result = m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_LIST) && (prop.itemType == SIP.SI_INT32));

			// Add the scan modes to the scan mode combo box
			scanModeComboBox.Items.Clear();
			for (int i = 0; i < prop.list.numItems; i++)
			{
				if(prop.list.items.piVal[i] == SIP.SI_SCANMODE_BW)
				{
					scanModeComboBox.Items.Add(m_blackAndWhiteText);
				}
				else if(prop.list.items.piVal[i] == SIP.SI_SCANMODE_GRAY)
				{
					scanModeComboBox.Items.Add(m_grayText);
				}
				else if(prop.list.items.piVal[i] == SIP.SI_SCANMODE_RGB)
				{
					scanModeComboBox.Items.Add(m_colorText);
				}

				if (prop.list.items.piVal[i] == prop.list.current.iVal)
				{
					scanModeComboBox.SelectedIndex = i;
				}
 			}

			// Although X resolution and Y resolution are separate properties,
			// scanners will typically have them locked together. X and Y will 
			// always be the same.
			prop.propertyID = SIP.SIP_XRESOLUTION;
			result = m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_LIST) && (prop.itemType == SIP.SI_INT32));

			// Add the resolutions to the resolution combobox.
			resolutionComboBox.Items.Clear();
			for (int i = 0; i < prop.list.numItems; i++)
			{
				resolutionComboBox.Items.Add(prop.list.items.piVal[i]);
			}

			resolutionComboBox.SelectedItem = prop.list.current.iVal;
		}



		private void CleanUp(bool disposing)
		{
			if (m_scannerReady && disposing)
			{
				m_scanner.SI_CloseInterface();
				m_scannerReady = false;
			}
		}

		// This function is called by the driver during calibration to 
		// report progress and allow the user to cancel.
		// Because of the way calibration works, there can be a few seconds 
		// between calls, so the GUI response may not be immediate.
		public SICallbackResult CalibrationProgressHandler(UInt32 percentComplete)
		{
			m_progressDialog.Percent = percentComplete;

			// This will process any pending messages, redrawing the progress 
			// meter and also responding to presses of the Cancel button.
			Application.DoEvents();

			if (m_progressDialog.Canceled)
			{	
				// Cancel the calibration
				return SICallbackResult.CANCEL;
			}

			return SICallbackResult.CONTINUE;

		}

		// Returns false if error.
		// Return true if:
		//	The scanner is already calibrated.
		//	The scanner performs the calibration.
		//	The user elected to skip calibration.
		//	(In this sample program, you can scan without calibrating.)
		private bool CheckForCalibration()
		{

			SICalibrationState calState = SICalibrationState.CALIBRATED;

			SIResult result;

			result = m_scanner.SI_IsCalibrated(ref calState);

			if (result == SIResult.SIR_SUCCESS)
			{
				if (calState == SICalibrationState.NOT_CALIBRATED)
				{

					DialogResult dlgRes = MessageBox.Show(
						"The scanner is not calibrated.\nDo you want to calibrate now?",
						"ScanToFile",
						MessageBoxButtons.YesNo,
						MessageBoxIcon.Exclamation);


					if (dlgRes == DialogResult.Yes)
					{
						if (!ConfirmPaper("Please insert the calibration paper and press OK.\n" +
								"Press Cancel to skip calibration."))
						{
							return true;
						}

						m_progressDialog = new ProgressDialog();
						m_progressDialog.Text = "Calibrating...";
						m_progressDialog.Show();
						Application.DoEvents();	// Allows the modeless dialog to fully draw itself

						result = m_scanner.SI_Calibrate(SICalibrationTarget.SI_CT_DEFAULT,
							new ProgressCallbackDelegate(CalibrationProgressHandler));

						m_progressDialog.Close();
						m_progressDialog = null;

						if (result == SIResult.SIR_SUCCESS)
						{
							MessageBox.Show("Calibration was successful.", "ScanToFile");
						}
						else
						{
							string errStr = "";
							m_scanner.SI_GetLastErrorText(ref errStr);
							MessageBox.Show(errStr, "ScanToFile",
								MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

							return false;
						}
					}

				}
			}
			else
			{
				return false;	// error
			}


			return true;

		}

		// returns true when paper has been inserted and the user clicks OK.
		// returns false if the user wants to abort the scan.
		bool ConfirmPaper(string message)
		{
			SIPaperStatus paperStatus = SIPaperStatus.SI_PS_PAPER_OUT;

			SIResult result = m_scanner.SI_GetPaperStatus(ref paperStatus);
			Debug.Assert(result == SIResult.SIR_SUCCESS);

			while (paperStatus == SIPaperStatus.SI_PS_PAPER_OUT)
			{
				DialogResult dlgRes = MessageBox.Show( 
					message,
					"ScanToFile",
					MessageBoxButtons.OKCancel);
				if (dlgRes == DialogResult.Cancel)
				{
					return false;	// skip 
				}
				result = m_scanner.SI_GetPaperStatus(ref paperStatus);
				Debug.Assert(result == SIResult.SIR_SUCCESS);
			}
			return true;
		}


		// This function retrives the current settings from the comboboxes and 
		// sets them to the scanner.
		void ControlsToProperties()
		{

			SIResult result;
			SIProperty prop = new SIProperty();

			////////////////////////////////////////////////
			// Scan Mode
			//////////////////////////////////////////
			prop.propertyID = SIP.SIP_SCAN_MODE;
			m_scanner.SI_GetProperty(prop);

			string str = (string)scanModeComboBox.SelectedItem;

			if (str.Equals(m_blackAndWhiteText))
			{
				prop.list.current.iVal = SIP.SI_SCANMODE_BW;
			}
			else if (str.Equals(m_grayText))
			{
				prop.list.current.iVal = SIP.SI_SCANMODE_GRAY;
			}
			else if (str.Equals(m_colorText))
			{
				prop.list.current.iVal = SIP.SI_SCANMODE_RGB;
			}
			
			result = m_scanner.SI_SetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);

			////////////////////////////////////////////////
			// Resolution
			//////////////////////////////////////////
			prop.propertyID = SIP.SIP_XRESOLUTION;
			m_scanner.SI_GetProperty(prop);

			prop.list.current.iVal = (int)resolutionComboBox.SelectedItem;

			result = m_scanner.SI_SetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);


			/////////////////////////////////////////////////////////////////
			// Scan width
			// To keep things simple, set the scan width to the default 
			// value for this resolution. Although, you can set any value 
			// between prop.range.minimum.iVal and prop.range.maximum.iVal.
			/////////////////////////////////////////////////////////////////
			prop.propertyID = SIP.SIP_SCAN_WIDTH_IN_PIXELS;
			m_scanner.SI_GetProperty(prop);
			prop.range.current.iVal = prop.range.def.iVal;

			result = m_scanner.SI_SetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);

			/////////////////////////////////////////////////////////////////
			// Scan length
			// To keep things simple, set the scan length to the default 
			// value for this resolution. Although, you can set any value 
			// between prop.range.minimum.iVal and prop.range.maximum.iVal.
			/////////////////////////////////////////////////////////////////
			prop.propertyID = SIP.SIP_SCAN_LENGTH_IN_LINES;
			m_scanner.SI_GetProperty(prop);
			prop.range.current.iVal = prop.range.def.iVal;

			result = m_scanner.SI_SetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);

			////////////////////////////////////////////////
			// Channel Order
			//
			// The driver returns color data in the order R-G-B by default.
			// But Windows bitmap format requires B-G-R order.
			// So we need to specify B-G-R order from the scanner driver.
			//////////////////////////////////////////
			prop.propertyID = SIP.SIP_CHANNEL_ORDER;
			m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_LIST) && (prop.itemType == SIP.SI_INT32));

			prop.list.current.iVal = SIP.SI_CO_BGR;	// set the order to BGR

			result = m_scanner.SI_SetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);

		}

		BitmapFile InitializeBitmapFile()
		{

			BitmapFile bmf = new BitmapFile();
			
			SIResult result;
			SIProperty prop = new SIProperty();

			// Get the current scan width (in pixels) from the scanner.
			prop.propertyID = SIP.SIP_SCAN_WIDTH_IN_PIXELS;
			result = m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_RANGE) && (prop.itemType == SIP.SI_INT32));

			// Specify scan width to the file
			bmf.SetWidth(prop.range.current.iVal);

			// Get the current scan length (height) from the scanner.
			prop.propertyID = SIP.SIP_SCAN_LENGTH_IN_LINES;
			result = m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_RANGE) && (prop.itemType == SIP.SI_INT32));

			// Specify scan length to the file. It will be vertically reversed, 
			// so negate the length.
			bmf.SetHeight(-prop.range.current.iVal);

			// Get the current number of bits per pixel from the scanner.
			prop.propertyID = SIP.SIP_BITS_PER_PIXEL;
			result = m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_SINGLE) && (prop.itemType == SIP.SI_INT32));

			// Specify scan length to the file
			bmf.SetBitsPerPixel((ushort)prop.single.current.iVal);

			// Get the current resolution from the scanner.
			prop.propertyID = SIP.SIP_XRESOLUTION;
			result = m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_LIST) && (prop.itemType == SIP.SI_INT32));
			bmf.SetResolution((ushort)prop.list.current.iVal);

			return bmf;

		}


		// returns the number of bytes in the buffer after alignment.
		private int AlignInPlace(ref byte[] buffer, int numberOfLines, int bytesPerLine)
		{
			if ((bytesPerLine % 4) != 0)
			{
				int bytesPerLineAligned = bytesPerLine + ((4 - (bytesPerLine % 4)) % 4);
				int srcIndex = bytesPerLine * (numberOfLines - 1);
				int destIndex = bytesPerLineAligned * (numberOfLines - 1);

				while (srcIndex >= 0)
				{
					Array.Copy(buffer, srcIndex, buffer, destIndex, bytesPerLine);

					destIndex -= bytesPerLineAligned;
					srcIndex -= bytesPerLine;
				}

				return bytesPerLineAligned * numberOfLines;
			}

			return bytesPerLine * numberOfLines;
		}


		private void OnScanButtonClicked(object sender, EventArgs e)
		{
			// Disable the scan button during scanning
			ScanButton.Enabled = false;

			// Check if calibration is needed and if so, do it.
			// This only returns false if there was an error.
			if (!CheckForCalibration())
			{
				ScanButton.Enabled = true;
				return;
			}

			// Make sure the user has paper inserted into the scanner.
			if(!ConfirmPaper("Please insert the item to be scanned and press OK.\n" +
								"Press Cancel to skip the scan."))
			{
				ScanButton.Enabled = true;
				return;
			}

			// Read the values from the UI controls and set them to the scanner.
			ControlsToProperties();

			// InitializeBitmapFile will read some of the default properties
			// from the scanner to initialize the bitmap file header.
			BitmapFile file = InitializeBitmapFile();

			// Open the file.
			try
			{
				file.Open(m_outFilename);
			}
			catch(Exception ee)
			{
				MessageBox.Show("Error opening the output file:  " + m_outFilename + "\n\n"
					+ ee.ToString(), "ScanToFile");
				ScanButton.Enabled = true;
				return;
			}


			// Allocate a buffer big enough for 10 lines.
			// We can find the space needed for one line by querying the
			// SIP_LINE_WIDTH_IN_BYTES property.
			SIResult result;
			SIProperty prop = new SIProperty();

			prop.propertyID = SIP.SIP_LINE_WIDTH_IN_BYTES;
			result = m_scanner.SI_GetProperty(prop);
			Debug.Assert(result == SIResult.SIR_SUCCESS);
			Debug.Assert((prop.containerType == SIP.SICON_SINGLE) && (prop.itemType == SIP.SI_INT32));

			int widthInBytes = prop.single.current.iVal;

			// Allocate enough memory to hold the scan lines aligned on 32-bit boundaries.
			// Later we'll align all scan lines received to 32-bit boundaries because that's
			// what Windows Bitmaps require.
			int widthInBytesAligned = widthInBytes + ((4 - (widthInBytes % 4)) % 4);
			byte[] lineBuffer = new byte[widthInBytesAligned * 10];

			
			// Now we're set to go.
			// Start the scan
			result = m_scanner.SI_StartScan();
			Debug.Assert(result == SIResult.SIR_SUCCESS);

			uint numLinesReturned;
			do
			{
				numLinesReturned = 0;

				// Attempt to read as many as 10 lines. We may get fewer than 
				// that if we're caught up to the scan.
				result = m_scanner.SI_ReadImageData(lineBuffer, 10, 0, ref numLinesReturned);

				// Align lines to 32-bit boundaries as required by Windows bitmap format 
				AlignInPlace(ref lineBuffer, (int)numLinesReturned, widthInBytes);

				// Write the data out to the file.
				if (result == SIResult.SIR_SUCCESS)
				{
					if (numLinesReturned > 0)
					{
						file.Write(lineBuffer, (uint)(numLinesReturned * widthInBytesAligned));
					}
				}

			} while (result == SIResult.SIR_SUCCESS);

			if (result == SIResult.SIR_ENDOFDATA)
			{
				// The full scan was done and we reached the end.
			}
			else if (result != SIResult.SIR_SUCCESS)
			{
				// error
				string errStr = "";
				m_scanner.SI_GetLastErrorText(ref errStr);
				MessageBox.Show(errStr, "ScanToFile");
			}

			result = m_scanner.SI_EndScan();
			Debug.Assert(result == SIResult.SIR_SUCCESS);

			// Feed the paper the rest of the way out of the scanner.
			m_scanner.SI_FeedPaperOut();

			try
			{
				file.Close();
			}
			catch (Exception ee)
			{
				MessageBox.Show(ee.ToString(), "File Exception");
				ScanButton.Enabled = true;
				return;
			}


			// Launch the file we just scanned.
			System.Diagnostics.ProcessStartInfo info = new System.Diagnostics.ProcessStartInfo(m_outFilename);
			info.UseShellExecute = true;
			info.Verb = "open";
			System.Diagnostics.Process.Start(info);

			ScanButton.Enabled = true;

		}
	}
}