//	File: TiffDump.cpp
//

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <assert.h>

#include "TiffDump.h"

const uint16 kTiffLittleEndian = 0x4949;	// "II"
const uint16 kTiffBigEndian = 0x4D4D;		// "MM"

const uint16 kTiffMagic = 42;			// "magic number" in header

// Type field values
const uint16 kTiffShortType =		3;
const uint16 kTiffLongType =		4;
const uint16 kTiffRationalType =	5;

// Tag field values
const uint16 kTiffWidthTag =			0x0100;
const uint16 kTiffLengthTag =			0x0101;
const uint16 kTiffBitsPerSampleTag =	0x0102;
const uint16 kTiffCompressionTag =		0x0103;
const uint16 kTiffPhotometricInterpTag = 0x0106;
const uint16 kTiffStripOffsetsTag =		0x0111;
const uint16 kTiffSamplesPerPixelTag =	0x0115;	// RGB only
const uint16 kTiffRowsPerStripTag =		0x0116;
const uint16 kTiffStripByteCountsTag =	0x0117;
const uint16 kTiffXResolutionTag =		0x011a;
const uint16 kTiffYResolutionTag =		0x011b;
const uint16 kTiffResolutionUnitTag =	0x0128;

const int kTiffMaxEntries = 12;		// max number of IFD entries

const int32 kBitsPerByte = 8;	// 'nuf said


// set struct alignment to 1 byte
#if defined(_WIN32) || defined(WIN32) || defined(_WINDOWS)
	#define PACK_STRUCT
	#pragma pack(push,1) // byte align structures
#elif defined(macintosh)
	#define PACK_STRUCT
	#pragma align_array_members off
	#pragma options align=mac68k
#elif defined(linux)
	#define PACK_STRUCT __attribute__((packed))
#endif

struct TiffEntry
{
	uint16	tag;
	uint16	type;
	uint32	count;
	union
	{
		uint16	sv;
		uint32	lv;

	} PACK_STRUCT;

} PACK_STRUCT;

struct TiffRational
{
	uint32	num;
	uint32	den;

} PACK_STRUCT;

struct TiffHeader
{
	uint16	byteOrder;
	uint16	magic;
	uint32	ifdOffset;

	char	fileIDstr[8];

	// Image File Directory
	uint16		numEntries;
	TiffEntry	entry[kTiffMaxEntries];
	uint32		next;

	// Values longer than four bytes
	TiffRational xResolution;
	TiffRational yResolution;
	uint16	bitsPerSample[4];

} PACK_STRUCT;

// restore struct alignment
#if defined(_WIN32) || defined(WIN32) || defined(_WINDOWS)
	#pragma pack(pop)
#elif defined(macintosh)
	#pragma options align=reset
	#pragma align_array_members reset
#endif
#undef PACK_STRUCT


//========================================================
// IsLittleEndian - Check processor endian-ness at run-time

inline bool IsLittleEndian()
{
	const int one = 1;	//c++ guarantees sizeof(int) >= 2
	return (*(uint8*)&one) != 0;
}

// ========================================================
// SetTiffEntry - Stuff values in a TIFF IFD entry.

void SetTiffEntry(uint16 tag, uint16 type, uint32 count, uint32 val, TiffEntry *ep)
{
	ep->tag = tag;
	ep->type = type;
	ep->count = count;
	if (type == kTiffShortType) {
		if (count == 1)
			ep->sv = (unsigned short) val;
		else
			ep->lv = val;	// 2 shorts or a file offset
	}
	else
		ep->lv = val;	// a long or a file offset
}

//========================================================

CTiffDump::CTiffDump() :
	fileRef(0),
	numLines(0),
	rowBytes(0),
	pixelWidth(0),
	numChannels(0),
	bitsPerChannel(0),
	resolution(72)
{
}


//========================================================

CTiffDump::~CTiffDump()
{	Cleanup();
}


//========================================================

void CTiffDump::Cleanup()
{	CloseFile();
}


//========================================================

void CTiffDump::CloseFile()
{
	if (fileRef != 0)
	{
		// If any lines were written, write the header
		// now that we know the final length
		if (numLines > 0)
			WriteHeader();

		// Close file
		fclose(fileRef);
		fileRef = 0;
	}
}


//========================================================

int CTiffDump::WriteHeader()
{
	TiffHeader	hdr;
	TiffEntry	*ep;
	int			wrCnt;
	int			numEntries;
	int			qq;

	// Make sure file is open
	if (fileRef == 0)
		return kTDWriteErr;

	// Stuff header

	memset(&hdr, 0, sizeof(hdr));	// Make sure pad bytes and IFD next field are zeroed

	if (numLines != 0)
	{
		ep = &hdr.entry[0];

		if (numChannels == 3)
		{
			// 24- or 48-bit RGB image

			if (bitsPerChannel == 8 || bitsPerChannel == 16)
			{
				SetTiffEntry(kTiffWidthTag, kTiffLongType, 1, pixelWidth, ep++);

				SetTiffEntry(kTiffLengthTag, kTiffLongType, 1, numLines, ep++);

				SetTiffEntry(kTiffBitsPerSampleTag, kTiffShortType, numChannels, offsetof(TiffHeader, bitsPerSample), ep++);

				SetTiffEntry(kTiffCompressionTag, kTiffShortType, 1, 1, ep++);					// 1 = uncompressed, 32773 = PackBits compression

				SetTiffEntry(kTiffPhotometricInterpTag, kTiffShortType, 1, 2, ep++);			// 2 = RGB

				SetTiffEntry(kTiffStripOffsetsTag, kTiffLongType, 1, sizeof(TiffHeader), ep++);	// Single strip follows header

				SetTiffEntry(kTiffSamplesPerPixelTag, kTiffShortType, 1, numChannels, ep++);	// Must be >= 3

				SetTiffEntry(kTiffRowsPerStripTag, kTiffLongType, 1, numLines, ep++);			// Single strip contains all rows

				SetTiffEntry(kTiffStripByteCountsTag, kTiffLongType, 1, numLines * rowBytes, ep++); // Single strip contains all bytes

				SetTiffEntry(kTiffXResolutionTag, kTiffRationalType, 1, offsetof(TiffHeader, xResolution), ep++);

				SetTiffEntry(kTiffYResolutionTag, kTiffRationalType, 1, offsetof(TiffHeader, yResolution), ep++);

				SetTiffEntry(kTiffResolutionUnitTag, kTiffShortType, 1, 2, ep++);				// 1 = none, 2 = inch, 3 = cm

				for (qq = 0; qq < numChannels; qq++)
					hdr.bitsPerSample[qq] = uint16(bitsPerChannel);
			}
			else
				return kTDParamErr;
		}
		else if (numChannels == 1)
		{
			// 8- or 16-bit gray or 1-bit B&W image

			if (bitsPerChannel == 8 || bitsPerChannel == 16 || bitsPerChannel == 1)
			{

				SetTiffEntry(kTiffWidthTag, kTiffLongType, 1, pixelWidth, ep++);

				SetTiffEntry(kTiffLengthTag, kTiffLongType, 1, numLines, ep++);

				SetTiffEntry(kTiffBitsPerSampleTag, kTiffShortType, numChannels, bitsPerChannel, ep++);

				SetTiffEntry(kTiffCompressionTag, kTiffShortType, 1, 1, ep++);					// 1 = uncompressed, 32773 = PackBits compression

				SetTiffEntry(kTiffPhotometricInterpTag, kTiffShortType, 1, 1, ep++);			// 0 = White is zero, 1 = Black is zero

				SetTiffEntry(kTiffStripOffsetsTag, kTiffLongType, 1, sizeof(TiffHeader), ep++);	// Single strip follows header

				SetTiffEntry(kTiffSamplesPerPixelTag, kTiffShortType, 1, numChannels, ep++);	// optional
			
				SetTiffEntry(kTiffRowsPerStripTag, kTiffLongType, 1, numLines, ep++);			// Single strip contains all rows

				SetTiffEntry(kTiffStripByteCountsTag, kTiffLongType, 1, numLines * rowBytes, ep++); // Single strip contains all bytes

				SetTiffEntry(kTiffXResolutionTag, kTiffRationalType, 1, offsetof(TiffHeader, xResolution), ep++);

				SetTiffEntry(kTiffYResolutionTag, kTiffRationalType, 1, offsetof(TiffHeader, yResolution), ep++);

				SetTiffEntry(kTiffResolutionUnitTag, kTiffShortType, 1, 2, ep++);				// 0 = none, 1 = inch, 3 = cm
			}
			else
				return kTDParamErr;
		}
		else
			return kTDParamErr;

		numEntries = int(ep - &hdr.entry[0]);
		assert(numEntries <= kTiffMaxEntries);

		if (numEntries != 0)
		{
			// Set fields common to all image types
			hdr.byteOrder = IsLittleEndian() ? kTiffLittleEndian : kTiffBigEndian;
			hdr.magic = kTiffMagic;
			hdr.ifdOffset = offsetof(TiffHeader, numEntries);

			strncpy(hdr.fileIDstr, "TiffDump", sizeof(hdr.fileIDstr));

			hdr.numEntries = uint16(numEntries);

			hdr.xResolution.num = resolution;
			hdr.xResolution.den = 1;

			hdr.yResolution.num = resolution;
			hdr.yResolution.den = 1;
		}
	}

	// Write it
	// rewind(fileRef);
	// rewind is preferrable here, but it isn't supported in Windows CE.
	clearerr(fileRef);
	fseek(fileRef, 0, SEEK_SET );

	wrCnt = (int)fwrite(&hdr, sizeof(hdr), 1, fileRef);
	if (wrCnt != 1)
		return kTDWriteErr;

	return kTDNoErr;
}

// ========================================================
// Create - create the image file. createPath specifies the
//file to create. This must be called before AppendLines().

int CTiffDump::Create(const char *createPath)
{
	if (fileRef != 0)
		return kTDOpenErr;

	// Open file
	fileRef = fopen(createPath, "wb");
	if (fileRef == 0)
		return kTDOpenErr;

	numLines = 0;
	rowBytes = 0;
	pixelWidth = 0;
	numChannels = 0;
	bitsPerChannel = 0;

	return kTDNoErr;
}


// ========================================================
// Routines to set image attributes.

void CTiffDump::SetWidthInPixels(int32 pw)
{
	if (numLines == 0 && pw > 0)
		pixelWidth = pw;
}

// --------------------------------------------------------

void CTiffDump::SetNumChannels(int nc)
{
	if (numLines == 0 && (nc == 1 || nc == 3))
		numChannels = nc;
}

// --------------------------------------------------------

void CTiffDump::SetBitsPerChannel(int bpc)
{

	if (numLines == 0 && (bpc == 1 || bpc == 8 || bpc == 16))
		bitsPerChannel = bpc;
}

// --------------------------------------------------------

void CTiffDump::SetResolution(int dpi)
{
	if (numLines == 0 && dpi > 0)
		resolution = dpi;
}

// ========================================================
// GetRowBytes - Returns bytes per line.

int32 CTiffDump::GetRowBytes()
{
	return (pixelWidth * numChannels * bitsPerChannel + (kBitsPerByte - 1)) / kBitsPerByte;
}


// ========================================================
// AppendLines - Append lines (rows) to the image file.

int CTiffDump::AppendLines(int32 linesToAppend, const void *pLineBuf)
{
	int		wrCnt;
	int		err;

	if (linesToAppend != 0)
	{
		// Make sure file is open
		if (fileRef == 0)
			return kTDWriteErr;
	
		if (numLines == 0)
		{
			// This is the first call
	
			// Check required parameters
			if (pixelWidth == 0 || numChannels == 0 || bitsPerChannel == 0)
				return kTDParamErr;		// Parameter not set
	
			// Multi-channel images must have at least 8 bits per channel
			if (numChannels > 1 && bitsPerChannel < kBitsPerByte )
				return kTDParamErr;
	
			// Write dummy header to set file position for first row
			err = WriteHeader();
			if (err != kTDNoErr)
				return err;
	
			rowBytes = GetRowBytes();
		}

		// Write lines to file
		wrCnt = (int)fwrite(pLineBuf, rowBytes, linesToAppend, fileRef);
		if (wrCnt != linesToAppend)
			return kTDWriteErr;
	
		numLines += linesToAppend;
	}
	return kTDNoErr;
}



// ========================================================

#ifdef TIFFDUMP_TEST_APP

int main(int argc, const char* argv[])
{
	int numChannels = 3;	// default
	int bitsPerChan = 8;	// default

	// First parameter is number of channels (optional)
	if (argc > 1)
		numChannels = atoi(argv[1]);
	// Second parameter is bits/pixel (optional)
	if (argc > 2)
		bitsPerChan = atoi(argv[2]);

	// Validate parameters
	if (! ( (numChannels == 1 && (bitsPerChan == 16 || bitsPerChan == 8 || bitsPerChan == 1)) ||
			(numChannels == 3 && (bitsPerChan == 16 || bitsPerChan == 8)) ) )
	{
		printf("Bad parameter\n");
		return 1;
	}

	// Create file
	CTiffDump tifdmp;
	if (tifdmp.Create("test.tif") != 0)
	{	printf("tifdmp.Create() failed\n");
		return 1;
	}

	// Set image file attributes
	const int32 widthInPixels	= 401;
	const int32 lengthInLines	= 600;
	const int	dotsPerInch		= 100;
	tifdmp.SetWidthInPixels(widthInPixels);
	tifdmp.SetNumChannels(numChannels);
	tifdmp.SetBitsPerChannel(bitsPerChan);
	tifdmp.SetResolution(dotsPerInch);

	int32 rowBytes = tifdmp.GetRowBytes();

	// Allocate buffer
	int32 maxBufLines = 65536L / rowBytes;
	uint8* buf = new uint8[maxBufLines * rowBytes];
	if (buf == 0)
	{	printf("Unable to allocate buffer\n");
		return 1;
	}

	// Loop to write image data
	int32 linesRemaining = lengthInLines;
	while (linesRemaining != 0)
	{
		// Compute number of lines to write
		int32 bufLines = linesRemaining;
		if (bufLines > maxBufLines)
			bufLines = maxBufLines;

		// Fill buffer with pixel data
		int32 curLine = lengthInLines - linesRemaining;
		uint8* linePtr = buf;
		for (int32 line = curLine; line < curLine + bufLines; line++)
		{
			switch (bitsPerChan)
			{
			case 16:
				{
					uint16* pp = (uint16*)linePtr;
					if (numChannels == 3)
					{
						for (int32 pixel = 0; pixel < widthInPixels; pixel++)
						{
							*pp++ = uint16(pixel);
							*pp++ = uint16(line + pixel);
							*pp++ = uint16(line);
						}
					}
					else	// numChannels == 1
					{
						for (int32 pixel = 0; pixel < widthInPixels; pixel++)
							*pp++ = uint16(line + pixel);
					}
				}
				break;
			case 8:
				{
					uint8* pp = linePtr;
					if (numChannels == 3)
					{
						for (int32 pixel = 0; pixel < widthInPixels; pixel++)
						{
							*pp++ = uint8(pixel);
							*pp++ = uint8(line + pixel);
							*pp++ = uint8(line);
						}
					}
					else	// numChannels == 1
					{
						for (int32 pixel = 0; pixel < widthInPixels; pixel++)
							*pp++ = uint8(line + pixel);
					}
				}
				break;
			case 1:
				memset(linePtr, ~(255 << (line % 9)), rowBytes);
				break;

			}
			linePtr += rowBytes;
		}

		// Write lines to file
		if (tifdmp.AppendLines(bufLines, buf) != 0)
		{
			printf("tifdmp.AppendLines() failed\n");
			return 1;
		}
		linesRemaining -= bufLines;
	}
	tifdmp.Cleanup();

	return 0;
}

#endif //TIFFDUMP_TEST_APP
