#include "StdAfx.h"
#include ".\bitmapfile.h"
#include "assert.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

BitmapFile::BitmapFile(void)
	:m_isOpen(false),
	m_dirty(false),
	m_hFile(INVALID_HANDLE_VALUE),
	m_totalImageBytesWritten(0),
	m_dataWritten(false),
	m_currentFileOffset(0)
{
	ZeroMemory(&m_bmFileHeader, sizeof(m_bmFileHeader));
	ZeroMemory(&m_bmInfoHeader, sizeof(m_bmInfoHeader));

	m_bmFileHeader.bfType = 0x4D42;		// the bytes "BM" for bitmap

	m_bmInfoHeader.biSize = sizeof(m_bmInfoHeader);
	m_bmInfoHeader.biPlanes = 1;
	m_bmInfoHeader.biCompression = BI_RGB;

	// initialize color table
	for(int i = 0; i < 256; i++)
	{
		m_colorTable[i].rgbBlue = 
		m_colorTable[i].rgbGreen = 
		m_colorTable[i].rgbRed = (BYTE)i;
		m_colorTable[i].rgbReserved = 0;
	}

}

BitmapFile::~BitmapFile(void)
{
}

bool BitmapFile::SetWidth(int width)
{
	m_bmInfoHeader.biWidth = width;
	m_dirty = true;
	return true;
}

bool BitmapFile::SetHeight(int height)
{
	m_bmInfoHeader.biHeight = height;
	m_dirty = true;
	return true;
}

bool BitmapFile::SetBitsPerPixel(int bitsPerPixel)
{
	if((bitsPerPixel == 1) || (bitsPerPixel == 8) || (bitsPerPixel == 24))
	{
		m_bmInfoHeader.biBitCount = bitsPerPixel;
		m_dirty = true;
		return true;
	}

	return false;
}


bool BitmapFile::Open(const char* filepath)
{
	assert(filepath);

	// If a file is currently open, close it.
	if(m_isOpen)
	{
		Close();
		m_hFile = INVALID_HANDLE_VALUE;
		m_isOpen = false;
	}

	m_hFile = CreateFile(
				filepath,				// file name 
				GENERIC_WRITE,			// open for write 
				0,						// do not share 
				NULL,					// default security 
			    CREATE_ALWAYS,			// overwrite existing file
				FILE_ATTRIBUTE_NORMAL,	// normal file 
				NULL);					// no template 


	if (m_hFile == INVALID_HANDLE_VALUE) 
	{ 
		return false;
	}

	m_dataWritten = false;
	m_currentFileOffset = 0;
	m_totalImageBytesWritten = 0;
	m_isOpen = true;

	return true;
}

bool BitmapFile::Seek(int filePosition)
{
	if(m_isOpen)
	{
		DWORD currentFilePosition = SetFilePointer( 
			m_hFile,		// must have GENERIC_READ and/or GENERIC_WRITE 
			filePosition,
			NULL,			// m_hFile is not large enough to need this pointer 
			FILE_BEGIN);	// provides offset from beginning of the file
	}
	else
	{
		return false;
	}
	return true;
}

bool BitmapFile::Write(const char * pBuffer, unsigned int lengthInBytes)
{
	bool result = false;
	DWORD bytesWritten;
	
	if(!m_isOpen)
	{
		return result;
	}

	assert(m_hFile);
	assert(pBuffer);

	if(m_dataWritten == false)
	{
		DoFirstWrite();
		m_dataWritten = true;
	}

	if(WriteFile(m_hFile, pBuffer, lengthInBytes, &bytesWritten, NULL))
	{
		if(lengthInBytes == bytesWritten)
		{
			m_totalImageBytesWritten += bytesWritten;
			result = true;
		}
	}

	return result;
}

// When Write() is called for the first time and image data is written 
// for the first time, we commit to the scan parameters as they are set 
// so that we write out the appropriate color table. The color table 
// needs to go before the image data.
void BitmapFile::DoFirstWrite()
{
	// With gray or black & white we need to write out a color table.
	int colorTableSizeInBytes = 0;
	if(m_bmInfoHeader.biBitCount == 8)	// grayscale
	{
		m_colorTable[1].rgbBlue = 
		m_colorTable[1].rgbGreen = 
		m_colorTable[1].rgbRed = 1;
		colorTableSizeInBytes =  sizeof(RGBQUAD) * 256;
	}
	else if(m_bmInfoHeader.biBitCount == 1)	// B&W
	{
		m_colorTable[1].rgbBlue = 
		m_colorTable[1].rgbGreen = 
		m_colorTable[1].rgbRed = 0xFF;
		colorTableSizeInBytes =  sizeof(RGBQUAD) * 2;
	}
	
	// Store total header size for use at the end of the scan
	m_bmFileHeader.bfOffBits = sizeof(m_bmFileHeader) + sizeof(m_bmInfoHeader) + colorTableSizeInBytes;

	Seek(0);

	DWORD bytesWritten;

	WriteFile(m_hFile,(char*)&m_bmFileHeader, sizeof(m_bmFileHeader), &bytesWritten, NULL);
	WriteFile(m_hFile,(char*)&m_bmInfoHeader, sizeof(m_bmInfoHeader), &bytesWritten, NULL);
	if(colorTableSizeInBytes > 0)
	{
		WriteFile(m_hFile,(char*)&m_colorTable[0], colorTableSizeInBytes, &bytesWritten, NULL);
	}

	m_currentFileOffset = m_bmFileHeader.bfOffBits;

}

void BitmapFile::Close()
{
	if(!m_isOpen)
	{
		return;
	}

	// Update the total file size according to the actual number of bytes written.
	m_bmFileHeader.bfSize = m_bmFileHeader.bfOffBits + m_totalImageBytesWritten;
	m_bmInfoHeader.biSizeImage = m_totalImageBytesWritten;

	Seek(0);

	DWORD bytesWritten;

	WriteFile(m_hFile,(char*)&m_bmFileHeader, sizeof(m_bmFileHeader), &bytesWritten, NULL);
	WriteFile(m_hFile,(char*)&m_bmInfoHeader, sizeof(m_bmInfoHeader), &bytesWritten, NULL);


	CloseHandle(m_hFile);
	m_hFile = INVALID_HANDLE_VALUE;
	m_isOpen = false;
}

