Imports System.IO
Imports System.Runtime.InteropServices


Structure BITMAPINFOHEADER
	Public biSize As UInt32
	Public biWidth As Int32
	Public biHeight As Int32
	Public biPlanes As UInt16
	Public biBitCount As UInt16
	Public biCompression As UInt32
	Public biSizeImage As UInt32
	Public biXPelsPerMeter As Int32
	Public biYPelsPerMeter As Int32
	Public biClrUsed As UInt32
	Public biClrImportant As UInt32
End Structure

Structure BITMAPFILEHEADER
	Public bfType As UInt16
	Public bfSize As UInt32
	Public bfReserved1 As UInt16
	Public bfReserved2 As UInt16
	Public bfOffBits As UInt32
End Structure

Structure RGBQUAD
	Public rgbBlue As Byte
	Public rgbGreen As Byte
	Public rgbRed As Byte
	Public rgbReserved As Byte
End Structure

Public Class BitmapFile

	Dim m_isOpen As Boolean
	Dim m_currentFileOffset As Integer
	Dim m_totalImageBytesWritten As Integer
	Dim m_dataWritten As Boolean
	Dim m_binaryWriter As BinaryWriter

	Dim m_bmFileHeader As BITMAPFILEHEADER
	Dim m_bmInfoHeader As BITMAPINFOHEADER
	Dim m_colorTable() As RGBQUAD
	Const m_sizeOfBitmapFileHeader As Integer = 14
	Const m_sizeOfBitmapInfoHeader As Integer = 40

	' constants for the biCompression field of BITMAPINFOHEADER
	Const BI_RGB As UInt32 = 0
	Const BI_RLE8 As UInt32 = 1
	Const BI_RLE4 As UInt32 = 2
	Const BI_BITFIELDS As UInt32 = 3
	Const BI_JPEG As UInt32 = 4
	Const BI_PNG As UInt32 = 5

	Public Sub New()
		m_isOpen = False
		m_totalImageBytesWritten = 0
		m_dataWritten = False
		m_currentFileOffset = 0

		m_bmFileHeader.bfType = &H4D42		' the bytes "BM" for bitmap
		m_bmFileHeader.bfReserved1 = 0
		m_bmFileHeader.bfReserved2 = 0

		m_bmInfoHeader.biSize = m_sizeOfBitmapInfoHeader
		m_bmInfoHeader.biPlanes = 1
		m_bmInfoHeader.biCompression = BI_RGB
		m_bmInfoHeader.biXPelsPerMeter = 0
		m_bmInfoHeader.biYPelsPerMeter = 0
		m_bmInfoHeader.biClrUsed = 0
		m_bmInfoHeader.biClrImportant = 0

		' initialize color table
		ReDim m_colorTable(255)		' allocate enough for grayscale
		For i As Integer = 0 To 255
			m_colorTable(i).rgbBlue = i
			m_colorTable(i).rgbGreen = i
			m_colorTable(i).rgbRed = i
			m_colorTable(i).rgbReserved = 0
		Next
	End Sub

	Public Sub Open(ByVal filepath As String)

		' If a file is currently open, close it.
		If m_isOpen Then
			m_binaryWriter.Close()
			m_isOpen = False
		End If

		m_binaryWriter = New BinaryWriter(New FileStream(filepath, FileMode.Create, FileAccess.Write))

		m_dataWritten = False
		m_currentFileOffset = 0
		m_totalImageBytesWritten = 0
		m_isOpen = True

		Return

	End Sub

	Public Function IsOpen() As Boolean
		Return m_isOpen
	End Function

	Public Sub Close()

		If Not m_isOpen Then
			Return
		End If

		' Update the total file size according to the actual number of bytes written.
		m_bmFileHeader.bfSize = (m_bmFileHeader.bfOffBits + m_totalImageBytesWritten)
		m_bmInfoHeader.biSizeImage = m_totalImageBytesWritten

		Seek(0)
		WriteHeaders()

		m_binaryWriter.Close()
		m_isOpen = False

	End Sub


	Public Function SetWidth(ByVal width As Integer) As Boolean
		m_bmInfoHeader.biWidth = width
		Return True
	End Function

	Public Function SetHeight(ByVal height As Integer)
		m_bmInfoHeader.biHeight = height
		Return True
	End Function


	Public Function SetBitsPerPixel(ByVal bitsPerPixel As UShort) As Boolean
		If ((bitsPerPixel = 1) Or (bitsPerPixel = 8) Or (bitsPerPixel = 24)) Then

			m_bmInfoHeader.biBitCount = bitsPerPixel
			Return True
		End If
		Return False
	End Function


	Public Function Seek(ByVal filePosition As Integer) As Boolean

		If m_isOpen Then

			Try

				m_binaryWriter.Seek(filePosition, SeekOrigin.Begin)

			Catch

				Return False

			End Try
		Else
			Return False
		End If

		Return True

	End Function

	Private Function StructureToByteArray(ByVal aStruct As Object) As Byte()
		Dim buff(Marshal.SizeOf(aStruct.GetType())) As Byte	  'Create Buffer
		Dim handle As GCHandle = GCHandle.Alloc(buff, GCHandleType.Pinned) ' Hands off GC
		'Marshal the structure
		Marshal.StructureToPtr(aStruct, handle.AddrOfPinnedObject(), False)
		handle.Free()
		Return buff
	End Function


	' Although a little uglier, using BinaryWriter is faster than writing out a 
	' whole structure as a block when the structure is less than about 40 fields.
	Sub WriteHeaders()

		m_binaryWriter.Write(m_bmFileHeader.bfType)
		m_binaryWriter.Write(m_bmFileHeader.bfSize)
		m_binaryWriter.Write(m_bmFileHeader.bfReserved1)
		m_binaryWriter.Write(m_bmFileHeader.bfReserved2)
		m_binaryWriter.Write(m_bmFileHeader.bfOffBits)

		m_binaryWriter.Write(m_bmInfoHeader.biSize)
		m_binaryWriter.Write(m_bmInfoHeader.biWidth)
		m_binaryWriter.Write(m_bmInfoHeader.biHeight)
		m_binaryWriter.Write(m_bmInfoHeader.biPlanes)
		m_binaryWriter.Write(m_bmInfoHeader.biBitCount)
		m_binaryWriter.Write(m_bmInfoHeader.biCompression)
		m_binaryWriter.Write(m_bmInfoHeader.biSizeImage)
		m_binaryWriter.Write(m_bmInfoHeader.biXPelsPerMeter)
		m_binaryWriter.Write(m_bmInfoHeader.biYPelsPerMeter)
		m_binaryWriter.Write(m_bmInfoHeader.biClrUsed)
		m_binaryWriter.Write(m_bmInfoHeader.biClrImportant)
	End Sub


	' When Write() is called for the first time and image data is written 
	' for the first time, we commit to the scan parameters as they are set 
	' so that we write out the appropriate color table. The color table 
	' needs to go before the image data.
	Private Sub DoFirstWrite()

		Const RGBQUADSIZE As Integer = 4

		' With gray or black & white we need to write out a color table.
		Dim colorTableEntriesToWrite As Integer = 0
		Dim colorTableSizeInBytes As Integer = 0
		If m_bmInfoHeader.biBitCount = 8 Then	 ' grayscale

			m_colorTable(1).rgbBlue = 1
			m_colorTable(1).rgbGreen = 1
			m_colorTable(1).rgbRed = 1
			colorTableSizeInBytes = RGBQUADSIZE * 256
			colorTableEntriesToWrite = 256

		ElseIf m_bmInfoHeader.biBitCount = 1 Then	' B&W

			m_colorTable(1).rgbBlue = &HFF
			m_colorTable(1).rgbGreen = &HFF
			m_colorTable(1).rgbRed = &HFF
			colorTableSizeInBytes = RGBQUADSIZE * 2
			colorTableEntriesToWrite = 2

		End If

		' Store total header size for use at the end of the scan
		m_bmFileHeader.bfOffBits = m_sizeOfBitmapFileHeader + _
				 m_sizeOfBitmapInfoHeader + _
				 colorTableSizeInBytes

		Seek(0)

		Try
			WriteHeaders()
		Catch
			Return
		End Try

		' write out the color table if necessary
		For i As Integer = 0 To colorTableEntriesToWrite - 1
			m_binaryWriter.Write(m_colorTable(i).rgbBlue)
			m_binaryWriter.Write(m_colorTable(i).rgbGreen)
			m_binaryWriter.Write(m_colorTable(i).rgbRed)
			m_binaryWriter.Write(m_colorTable(i).rgbReserved)
		Next

		m_currentFileOffset = m_bmFileHeader.bfOffBits
	End Sub

	' Write a buffer out to the file.
	Public Function Write(ByVal buffer As Byte(), ByVal lengthInBytes As UInteger) As Boolean

		Dim result As Boolean = False

		If Not m_isOpen Then
			Return result
		End If

		If m_dataWritten = False Then

			DoFirstWrite()
			m_dataWritten = True
		End If

		Try
			m_binaryWriter.BaseStream.Write(buffer, 0, lengthInBytes)
		Catch
			Return False
		End Try

		m_totalImageBytesWritten += lengthInBytes
		result = True
		Return result
	End Function


End Class


