/** \file ScannerAPI.h
	The Scanner API functions exported by the DLL.
**/

#if !defined(SCANNERAPI_H)
#define SCANNERAPI_H

#include "PlatTypes.h"



typedef uint32 SIResult;

const SIResult SIR_SUCCESS =						0;
const SIResult SIR_UNKNOWN_MODEL_NAME =				10;
const SIResult SIR_SCANNER_NOT_READY =				11;
const SIResult SIR_BAD_PARAMETER =					14;
const SIResult SIR_ALREADY_OPEN =					15;
const SIResult SIR_INTERFACE_NOT_OPEN =				17;
const SIResult SIR_FILESYSTEM_ACCESS_DENIED =		18;
const SIResult SIR_SCANNER_BUSY =					19;
const SIResult SIR_ENDOFDATA =						20;	
const SIResult SIR_NOT_SCANNING =					21;	
const SIResult SIR_DEVICE_COMMUNICATION_ERROR =		0x100;
const SIResult SIR_UNKNOWN_ERROR =					0x800;
const SIResult SIR_NOT_CALIBRATED =					0x802;
const SIResult SIR_INITIALIZATION_FAILURE =			0x803;
const SIResult SIR_USER_CANCELLED =					0x804;
const SIResult SIR_CALIB_WHITE_TARGET =				0x840;
const SIResult SIR_PROPERTY_UNSUPPORTED =			0x900;	///< The caller tried to set or retrieve a capabiity that is not supported in the current implementation.
const SIResult SIR_PROPERTY_INVALID_VALUE =			0x901;	///< The caller tried to set a property that was outside of the valid range.
const SIResult SIR_PROPERTY_READONLY =				0x902;	///< The caller tried to set a property that has read-only access.
const SIResult SIR_END_OF_CODES =					0xFFFFFFFF;


#define MAX_EVENT_DATA_SIZE  20 
const uint32	SIEVT_BUTTON_DOWN =		0x00000001;
const uint32	SIEVT_BUTTON2_DOWN =	0x00000002;
const uint32	SIEVT_BUTTON3_DOWN =	0x00000004;
const uint32	SIEVT_PAPER_IN =		0x00000010;

/** \brief Used as a parameter in the SI_GetEvent() function.

	The SIEventInfo structure is used to pass event information to the 
	SI_GetEvent() API function. The caller must fill out the eventDataSize 
	and eventData[] fields. upon return, the function will fill out the 
	eventName[] field.
**/
typedef struct 
{
    uint8	eventDataSize;					///< [in] The size in bytes of the data in the eventData array.
    uint8	eventData[MAX_EVENT_DATA_SIZE];	///< [in] The data return by the scanner in the USB interrupt phase.
    uint32	eventFlags;						///< [out] Bit flags that indicate which event(s) occurred.
} SIEventInfo;

// Scanner status.
typedef uint32 SIScannerStatus;
const uint32 SI_SS_OFFLINE =		0;
const uint32 SI_SS_ONLINE_READY =	1;
const uint32 SI_SS_ONLINE_BUSY =	2;

// Paper status.
typedef uint32 SIPaperStatus;
const uint32 SI_PS_PAPER_OUT =		0;
const uint32 SI_PS_PAPER_IN =		1;

// Button status.
typedef uint32 SIButtonStatus;
const uint32 SI_BS_DOWN =		1;
const uint32 SI_BS_UP =			0;

// Diagnostic Tests
typedef uint32 SIDiagTest;
const uint32 SI_DIAG_HW_PRESENCE =		0x00;		///< Test for the presence of the scanner.
const uint32 SI_DIAG_READ_PERFORMANCE =	0x10;		///< Read a small sample of data from the scanner hardware to test the port speed.

// Diagnostic info structure for SI_DIAG_HW_PRESENCE. Passed as second parameter.
typedef struct {
	uint32 result;		///<	[out] Will be set to either SI_DIAG_PASS or SI_DIAG_FAIL.
} SIDiagHwPresenceInfo;

const uint32 SI_DIAG_FAIL =		0;		///< The diagnostic test failed.
const uint32 SI_DIAG_PASS =		1;		///< The diagnostic test passed.

// Diagnostic info structure for SI_DIAG_READ_PERFORMANCE. Passed as second parameter.
typedef struct {
	uint32 readLengthInKB;		///<	[in]  Set this to the length of data to read from the hardware.
	uint32 timeInMicroSec;		///<	[out] Resulting time in microseconds will be returned here.
} SIDiagReadPerfInfo;


typedef uint32 SICalibrationState;		///< Parameter type passed to SI_IsCalibrated().
										///< Set to SI_TRUE if the scanner is calibrated.
										///< Set to SI_FALSE if the scanner is not calibrated.

// Valid parameters for the SI_Calibrate() function.
// These are types of calibration targets.
const uint32 SI_CT_DEFAULT =	0;
const uint32 SI_CT_WHITEONLY =	1;

/** This is the maximum length (including NULL terminator) of any string 
	that will be returned from GetLastErrorText().	**/
const int SI_MAX_ERROR_STRING_LENGTH = 80;



///////////////////////////////////////////////////////////////////////////
//
//		Properties
//
///////////////////////////////////////////////////////////////////////////

// Set structure alignment to 8. 
#pragma PACK8


// Property IDs
const uint16 SIP_INVALID =						0;	///< NULL or invalid property
const uint16 SIP_OPTICAL_RESOLUTION =			2;	///< in pixels per inch
const uint16 SIP_OPTICAL_WIDTH_IN_PIXELS =		3;
const uint16 SIP_XRESOLUTION =					4;
const uint16 SIP_YRESOLUTION =					5;
const uint16 SIP_BITS_PER_CHANNEL =				6;
const uint16 SIP_BITS_PER_PIXEL =				7;
const uint16 SIP_SCAN_MODE =					8;
const uint16 SIP_BRIGHTNESS =					9;
const uint16 SIP_CONTRAST =						10;
const uint16 SIP_PHOTOMETRIC_INTERPRETATION =	11;
const uint16 SIP_YOFFSET =						12;
const uint16 SIP_XOFFSET =						13;
const uint16 SIP_SCAN_WIDTH_IN_PIXELS =			14;
const uint16 SIP_SCAN_LENGTH_IN_LINES =			15;
const uint16 SIP_PLANARCHUNKY =					16;
const uint16 SIP_CHANNEL_ORDER =				17;	///< The order of RGB bytes in a color pixel. valid values are constants that begin with SI_CO_
const uint16 SIP_MAX_SCAN_TIME_IN_SEC =			18;
const uint16 SIP_LINE_WIDTH_IN_BYTES =			19;	///< always a read-only value. Depends on scan mode, x offset, width.
const uint16 SIP_DUPLEX_ENABLED =				21;
const uint16 SIP_SPOOLER_ENABLED =				22;
const uint16 SIP_FEED_DIRECTION =				23;
const uint16 SIP_OPTICAL_LENGTH_IN_PIXELS =		24;
const uint16 SIP_GAMMA =						25;
const uint16 SIP_HIGHLIGHT =					26;
const uint16 SIP_SHADOW =						27;
const uint16 SIP_LUT_GRAY =						28;
const uint16 SIP_LUT_RED =						29;
const uint16 SIP_LUT_GREEN =					30;
const uint16 SIP_LUT_BLUE =						31;
const uint16 SIP_THRESHOLD =					32;
const uint16 SIP_EOP_DETECT_ENABLED =			33;
const uint16 SIP_DESCREEN_ENABLED =				34;
const uint16 SIP_DROPOUT_COLOR =				35;
const uint16 SIP_LED_INDICATOR1 =				40;
const uint16 SIP_LED_INDICATOR2 =				41;
const uint16 SIP_PREFEED_ENABLED =				50;
const uint16 SIP_PREFEED_DELAY =				51;
const uint16 SIP_PREFEED_DISTANCE =				52;
const uint16 SIP_EOP_DETECT_OFFSET =			60;
const uint16 SIP_USB_RATE =						220;
const uint16 SIP_USB_SERIAL_NUMBER =			225;
const uint16 SIP_FEED_RATE =					1004;
const uint16 SIP_SCAN_RATE =					1006;


const uint32 SI_SCANMODE_BW =	0;
const uint32 SI_SCANMODE_GRAY =	1;
const uint32 SI_SCANMODE_RGB =	2;

const uint32 SI_PI_BLACKZERO =	0;	///< Photometric Interpretation. Black is represented by zero
const uint32 SI_PI_WHITEZERO =	1;	///< Photometric Interpretation. White is represented by zero

const uint32 SI_PC_PLANAR =		0;
const uint32 SI_PC_CHUNKY =		1;	

const uint32 SI_CO_RGB =		0;	///< Value of SIP_CHANNEL_ORDER. SI_CO_RGB specifies that pixels are in red-green-blue order.
const uint32 SI_CO_BGR =		1;	///< Value of SIP_CHANNEL_ORDER. SI_CO_BRG specifies that pixels are in blue-green-red order.

const uint32 SI_DO_NONE =		0;	///< Value of SIP_DROPOUT_COLOR. SI_DO_NONE means no dropout.
const uint32 SI_DO_RED =		1;	///< Value of SIP_DROPOUT_COLOR. SI_DO_RED means red dropout.
const uint32 SI_DO_GREEN =		2;	///< Value of SIP_DROPOUT_COLOR. SI_DO_GREEN means green dropout.
const uint32 SI_DO_BLUE =		3;	///< Value of SIP_DROPOUT_COLOR. SI_DO_BLUE means blue dropout.

// Container types
const uint16 SICON_SINGLE =	0;
const uint16 SICON_RANGE =	1;
const uint16 SICON_LIST =	2;
const uint16 SICON_ARRAY =	3;

// Item types
const uint16 SI_INT32 =		1;		///< 32-bit signed integer
const uint16 SI_UINT32 =	2;		///< 32-bit unsigned integer
const uint16 SI_INT16 =		3;		///< 16-bit signed integer
const uint16 SI_UINT16 =	4;		///< 16-bit unsigned integer
const uint16 SI_INT8 =		5;		///< 8-bit signed integer
const uint16 SI_UINT8 =		6;		///< 8-bit unsigned integer
const uint16 SI_BOOL =		7;		///< The item will be either SI_TRUE or SI_FALSE.
const uint16 SI_FLOAT32 =	8;		///< single-precision floating point 
const uint16 SI_STR =		9;		///< string. The item will be a pointer to an ANSI zero-terminated string.

/**	

	When the type is specified as a SI_STR, the 32-bit value field holds a
	pointer to a null terminated ASCII string. \n \n
	For a list, the value points to a list of pointers to strings. The 
	current and default values also hold pointers to strings, and these pointers will 
	match one of the pointers in the list. Likewise, an array holds an array of pointers 
	to strings. Ranges do not hold strings. \n \n
	Memory allocation: string pointers will point to memory within the DLL. There's no 
	need for deallocation by the application. The application must not call delete on 
	a string pointer.
**/

const uint32 SI_FALSE =	0;
const uint32 SI_TRUE =	1;

// Motor feed directions. Used for SIP_FEED_DIRECTION.
const uint32 SI_FORWARD =	0;
const uint32 SI_REVERSE =	1;

// These identifiers used with the SIP_USB_RATE property.
const uint32 SI_USB_FULL =	1;
const uint32 SI_USB_HIGH =	2;

typedef struct 
{
	union
	{
		int32	iVal;		///< integer value (SI_INT32)
		float32	fVal;		///< float value (SI_FLOAT32)
		const char*	strVal;	///< string value (SI_STR)
		int32	bVal;		///< boolean value. (SI_BOOL. Either SI_TRUE or SI_FALSE)

		// The following are generally used only with arrays.
		int32*		piVal;		///< pointer to integer (SI_INT32)
		int16*		psVal;		///< pointer to short (SI_INT16)
		int8*		pcVal;		///< pointer to character (SI_INT8)
		uint32*		puiVal;		///< pointer to unsigned integer (SI_UINT32)
		uint16*		pusVal;		///< pointer to unsigned short (SI_UINT16)
		uint8*		pucVal;		///< pointer to unsigned character (SI_UINT8)
		char**		pstrVal;	///< pointer to strings (SI_STR)
		int32*		pbVal;		///< pointer to boolean (SI_BOOL)
		float32*	pfVal;		///< pointer to float (SI_FLOAT32)
	};
} SIValue;


typedef struct 
{
	SIValue	current;
	SIValue	def;
} SISingle;


typedef	struct
{
	SIValue	minimum;
	SIValue	maximum;
	SIValue	stepSize;
	SIValue	def;
	SIValue	current;
} SIRange;


typedef struct
{
	int32	numItems;
	SIValue	current;
	SIValue	def;
	SIValue	items;
} SIList;



typedef struct
{
	int32	numItems;
	SIValue	items;
} SIArray;


const uint16 SIACC_READONLY =	0;
const uint16 SIACC_READWRITE =	1;

typedef struct
{
	uint16		propertyID;			///< a property ID constant.
	uint16		containerType;		///< a container type constant.
	uint16		itemType;			///< the data type the container holds.
	uint16		access;				///< Set to SIACC_READWRITE if this property can be changed. 
									///< Set to SIACC_READONLY if this property is read-only.
	// containers
	union
	{
		SISingle	single;
		SIRange		range;
		SIList		list;
		SIArray		array;
	};
} SIProperty;

// Restore structure alignment to previous value
#pragma UNPACK

///////////////////////////////////////////////////////////////////////////

typedef int32 (SICALLBACK *ProgressCallbackPtr)(uint32 percentCompleted);

///////////////////////////////////////////////////////////////////////////


#if defined(__cplusplus)
extern "C" 
{
#endif

	SCANNERDLL_API SIResult SICALLTYPE	SI_OpenInterface(const char* pName);
	SCANNERDLL_API SIResult SICALLTYPE	SI_CloseInterface();

	SCANNERDLL_API SIResult SICALLTYPE	SI_IsCalibrated(SICalibrationState *pState);
	SCANNERDLL_API SIResult SICALLTYPE	SI_Calibrate(uint32 target, ProgressCallbackPtr pProgressCallback);
	SCANNERDLL_API SIResult SICALLTYPE	SI_GetScannerStatus(SIScannerStatus *pScannerStatus);
	SCANNERDLL_API SIResult SICALLTYPE	SI_GetPaperStatus(SIPaperStatus *pPaperStatus);
	SCANNERDLL_API SIResult SICALLTYPE	SI_GetButtonStatus(uint32 buttonNumber, SIButtonStatus *pButtonStatus);

	SCANNERDLL_API SIResult SICALLTYPE	SI_GetProperty(SIProperty* pProperty);
	SCANNERDLL_API SIResult SICALLTYPE	SI_SetProperty(SIProperty* pProperty);

	SCANNERDLL_API SIResult SICALLTYPE	SI_StartScan();
	SCANNERDLL_API SIResult SICALLTYPE	SI_ReadImageData(uint8 *buffer, uint32 numberOfLinesToRead, uint32 pageNumber, uint32* numberOfLinesReturned );
	SCANNERDLL_API SIResult SICALLTYPE	SI_EndScan();
	SCANNERDLL_API SIResult SICALLTYPE	SI_Feed();
	SCANNERDLL_API SIResult SICALLTYPE	SI_FeedPaperOut();
	SCANNERDLL_API SIResult SICALLTYPE	SI_Clean(ProgressCallbackPtr pProgressCallback);

	SCANNERDLL_API SIResult SICALLTYPE	SI_Reset();
	SCANNERDLL_API SIResult SICALLTYPE	SI_Diagnostic(SIDiagTest test, void* pDiagInfo);
	SCANNERDLL_API SIResult SICALLTYPE	SI_GetLastError();
	SCANNERDLL_API SIResult SICALLTYPE	SI_GetLastErrorText(char **ppErrorText);
	SCANNERDLL_API SIResult SICALLTYPE	SI_GetEvent(SIEventInfo* eventInfo);

#if defined(__cplusplus)
}
#endif


#endif
